"""
Tests for filters in enterprise_data.
"""

from unittest import mock
from urllib.parse import urljoin

import responses
from rest_framework import status
from rest_framework.reverse import reverse
from rest_framework.test import APIRequestFactory, APITestCase

from django.conf import settings

from enterprise_data.api.v1.views import EnterpriseLearnerViewSet
from enterprise_data.filters import AuditUsersEnrollmentFilterBackend
from enterprise_data.models import EnterpriseEnrollment, EnterpriseLearnerEnrollment
from enterprise_data.tests.mixins import JWTTestMixin
from enterprise_data.tests.test_utils import EnterpriseLearnerEnrollmentFactory, EnterpriseLearnerFactory, UserFactory


class TestConsentGrantedFilterBackend(JWTTestMixin, APITestCase):
    """
    Test suite for the ``TestConsentGrantedFilterBackend`` filter.
    """
    fixtures = ('enterprise_enrollment', 'enterprise_user', )

    def setUp(self):
        super().setUp()
        self.user = UserFactory(is_staff=True)
        self.client.force_authenticate(user=self.user)
        self.enterprise_id = 'ee5e6b3a-069a-4947-bb8d-d2dbc323396c'
        self.url = reverse('v0:enterprise-enrollments-list',
                           kwargs={'enterprise_id': self.enterprise_id})
        self.set_jwt_cookie(context=self.enterprise_id)

    def mock_enterprise_api_client_response(self, enforce_dsc=None):
        """
        Mocks enterprise api client response.
        """
        mock_path = 'enterprise_data.filters.EnterpriseApiClient.get_enterprise_customer'
        with mock.patch(mock_path) as mock_enterprise_api_client:
            mock_enterprise_api_client.return_value = {
                'uuid': self.enterprise_id,
                'enable_audit_data_reporting': True,
                'enforce_data_sharing_consent': enforce_dsc or True,
            }
            return self.client.get(self.url)

    def test_filter_for_list(self):
        """
        Filter users through email/username if requesting user is staff, otherwise based off of request user ID.
        """
        response = self.mock_enterprise_api_client_response()

        assert response.status_code == status.HTTP_200_OK
        data = response.json()

        # Fixture data for enterprise 'ee5e6b3a-069a-4947-bb8d-d2dbc323396c' contains 3 objects but only 2 with consent
        assert EnterpriseEnrollment.objects.filter(enterprise_id=self.enterprise_id).count() == 3
        assert len(data['results']) == 2

    def test_filter_with_external_consent(self):
        """
        If the enterprise is configured for externally managed data consent, all enrollments will be returned.
        """
        response = self.mock_enterprise_api_client_response('externally_managed')
        assert response.status_code == status.HTTP_200_OK
        data = response.json()

        # Fixture data for enterprise 'ee5e6b3a-069a-4947-bb8d-d2dbc323396c' contains 3 objects but only 2 with consent
        assert EnterpriseEnrollment.objects.filter(enterprise_id=self.enterprise_id).count() == 3
        assert len(data['results']) == 3


class TestAuditUsersEnrollmentFilterBackend(JWTTestMixin, APITestCase):
    """
    Test suite for the ``AuditUsersEnrollmentFilterBackend``.
    """

    def setUp(self):
        super().setUp()
        self.user = UserFactory(is_staff=True)
        self.client.force_authenticate(user=self.user)
        self.enterprise1_uuid = 'ee5e6b3a-069a-4947-bb8d-d2dbc323396c'
        self.enterprise2_uuid = '0395b02f-6b29-42ed-9a41-45f3dff8349c'
        self.set_jwt_cookie(context=self.enterprise1_uuid)
        self.set_jwt_cookie(context=self.enterprise2_uuid)

    def mock_enterprise_api_endpoints(self, enterprise_uuid, enable_audit_data_reporting=True):
        """
        Mock enterprise api endpoints.
        """
        responses.add(
            responses.POST,
            urljoin(settings.BACKEND_SERVICE_EDX_OAUTH2_PROVIDER_URL + '/', 'access_token'),
            status=200,
            json={
                'access_token': 'test_access_token',
                'expires_in': 3600,
            },
            content_type='application/json'
        )

        responses.add(
            responses.GET,
            urljoin(settings.LMS_BASE_URL + '/', f'enterprise/api/v1/enterprise-customer/{enterprise_uuid}'),
            json={
                'uuid': enterprise_uuid,
                'enable_audit_data_reporting': enable_audit_data_reporting,
                'enforce_data_sharing_consent': True,
            },
            status=200,
            content_type='application/json'
        )

    @responses.activate
    def test_filter_with_audit_reporting_enabled(self):
        """
        Verify that expected response is returned when `enable_audit_data_reporting` set to True.
        """
        enterprise_uuid = self.enterprise1_uuid
        enterprise_learner = EnterpriseLearnerFactory(enterprise_customer_uuid=enterprise_uuid)
        EnterpriseLearnerEnrollmentFactory.create_batch(
            2,
            is_consent_granted=True,
            enterprise_user_id=enterprise_learner.enterprise_user_id
        )

        queryset = EnterpriseLearnerEnrollment.objects.all()
        request = APIRequestFactory().get('/')
        view = EnterpriseLearnerViewSet(kwargs={'enterprise_id': enterprise_uuid})
        filter_backend = AuditUsersEnrollmentFilterBackend()

        self.mock_enterprise_api_endpoints(enterprise_uuid=enterprise_uuid)

        filtered_queryset = filter_backend.filter_queryset(request, queryset, view)
        assert filtered_queryset == queryset

    @responses.activate
    def test_filter_without_audit_reporting_enabled(self):
        """
        Verify that expected response is returned when `enable_audit_data_reporting` set to False.
        """
        enterprise_uuid = self.enterprise2_uuid
        enterprise_learner_1 = EnterpriseLearnerFactory(enterprise_customer_uuid=enterprise_uuid)
        learner_enrollment_1 = EnterpriseLearnerEnrollmentFactory(
            enterprise_customer_uuid=enterprise_uuid,
            is_consent_granted=True,
            enterprise_user_id=enterprise_learner_1.enterprise_user_id
        )

        enterprise_learner_2 = EnterpriseLearnerFactory(enterprise_customer_uuid=enterprise_uuid)
        EnterpriseLearnerEnrollmentFactory(
            enterprise_customer_uuid=enterprise_uuid,
            is_consent_granted=True,
            user_current_enrollment_mode='audit',
            enterprise_user_id=enterprise_learner_2.enterprise_user_id
        )

        queryset = EnterpriseLearnerEnrollment.objects.all()
        request = APIRequestFactory().get('/')
        view = EnterpriseLearnerViewSet(kwargs={'enterprise_id': enterprise_uuid})
        filter_backend = AuditUsersEnrollmentFilterBackend()

        self.mock_enterprise_api_endpoints(enterprise_uuid=enterprise_uuid, enable_audit_data_reporting=False)

        filtered_queryset = filter_backend.filter_queryset(request, queryset, view)
        assert filtered_queryset.count() == 1
        assert filtered_queryset.first().enterprise_enrollment_id == learner_enrollment_1.enterprise_enrollment_id
        assert filtered_queryset.first().user_current_enrollment_mode != 'audit'



"""
Test reporter.
"""
import datetime
import os
import unittest
from unittest import mock
from unittest.mock import MagicMock

import ddt
import pytest

from enterprise_reporting import reporter
from enterprise_reporting.reporter import EnterpriseReportSender
from enterprise_reporting.utils import encrypt_string


@ddt.ddt
class TestReporter(unittest.TestCase):
	""""
	Tests about reporter methods
	"""
	def setUp(self):
		self.FILE_WRITE_DIRECTORY = '/tmp'
		self.date = datetime.datetime.now().strftime("%Y-%m-%d")
		self.reporting_config = {
			'enterprise_customer': {
				'uuid': '12aacfee8ffa4cb3bed1059565a57f05', 'name': 'John'
			},
			'delivery_method': 'sftp',
			'data_type': 'catalog',
			'uuid': '987654321',
			'sftp_hostname': '127.0.0.1',
			'active': True,
			'report_type': 'json',
			'encrypted_password': 'bleeH-bLah-blUe001',
			'sftp_username': 'John1',
			'sftp_port': 22,
			'sftp_file_path': 'home/user/Documents',
			'encrypted_sftp_password': 'abraAcaDabradOo',
		}
		sftp_password = self.reporting_config['encrypted_sftp_password']
		self.reporting_config['encrypted_sftp_password'] = encrypt_string(sftp_password)

	@ddt.data(
		True,
		False
	)
	def test_data_report_file_name(self, flag_value):
		"""
		Tests if the report config file name is generated correctly when date is included
		"""
		self.reporting_config['include_date'] = flag_value
		enterprise_report_sender = EnterpriseReportSender.create(self.reporting_config)
		actual_file_name = enterprise_report_sender.data_report_file_name
		date_str = f"_{self.date}" if self.reporting_config.get('include_date') else ""
		expected_file_name = "{dir}/{enterprise_uuid}_{data_type}_{report_type}{date}.{report_type}".format(
			dir=self.FILE_WRITE_DIRECTORY,
			enterprise_uuid=self.reporting_config['enterprise_customer']['uuid'],
			data_type=self.reporting_config['data_type'],
			date=date_str,
			report_type=self.reporting_config['report_type']
		)
		assert actual_file_name == expected_file_name

	@ddt.data(
		'grade',
		'course_structure',
		'completion',
	)
	@mock.patch("enterprise_reporting.reporter.S3Client")
	def test_manual_reports(self, data_type, mock_s3_client):
		"""
		Verify that `get_enterprise_report` has been called with correct S3 CSV path.
		"""
		# mock S3Client get_enterprise_report to verify if it has called with correct arguments
		mock_s3_client.return_value.get_enterprise_report.return_value = MagicMock()

		report_config = dict(self.reporting_config, data_type=data_type, report_type="csv")
		enterprise_uuid = report_config['enterprise_customer']['uuid']

		# set environment variable
		s3_csv_path = f"BATMAN/REPORTS/{data_type}.csv"
		os.environ[f"{data_type}-{enterprise_uuid}"] = s3_csv_path

		enterprise_report_sender = EnterpriseReportSender.create(report_config)
		data_report_file = getattr(enterprise_report_sender, f'_generate_enterprise_report_{data_type}_csv')()

		mock_s3_client.return_value.get_enterprise_report.assert_called_with(s3_csv_path, data_report_file[0])

	@ddt.data(
		'grade',
		'course_structure',
		'completion',
	)
	@mock.patch("enterprise_reporting.reporter.S3Client")
	def test_manual_reports_without_env_var(self, data_type, mock_s3_client):
		"""
		Verify that `get_enterprise_report` raises correct exception if S3 csv path is not present as an env variable.
		"""
		# mock S3Client get_enterprise_report to verify if it has not called
		mock_s3_client.return_value.get_enterprise_report.return_value = MagicMock()

		report_config = dict(self.reporting_config, data_type=data_type, report_type="csv")
		enterprise_uuid = report_config['enterprise_customer']['uuid']

		# remove environment variable
		del os.environ[f"{data_type}-{enterprise_uuid}"]

		enterprise_report_sender = EnterpriseReportSender.create(report_config)
		with pytest.raises(ValueError) as ve:
			getattr(enterprise_report_sender, f'_generate_enterprise_report_{data_type}_csv')()

		assert ve.value.args[0] == 'Invalid S3 CSV path. Path: [None]'
		mock_s3_client.return_value.get_enterprise_report.assert_not_called()



"""
Tests for views in the `enterprise_data` module.
"""

import datetime
import os
from unittest import mock
from uuid import UUID, uuid4

import ddt
from pytest import mark
from rest_framework import status
from rest_framework.reverse import reverse
from rest_framework.test import APITransactionTestCase

from enterprise_data.api.v1.serializers import EnterpriseOfferSerializer
from enterprise_data.models import EnterpriseLearnerEnrollment, EnterpriseOffer
from enterprise_data.tests.factories import (
    EnterpriseAdminLearnerProgressFactory,
    EnterpriseAdminSummarizeInsightsFactory,
)
from enterprise_data.tests.mixins import JWTTestMixin
from enterprise_data.tests.test_utils import (
    EnterpriseLearnerEnrollmentFactory,
    EnterpriseLearnerFactory,
    EnterpriseOfferFactory,
    UserFactory,
    get_dummy_enterprise_api_data,
)
from enterprise_data_roles.constants import ENTERPRISE_DATA_ADMIN_ROLE
from enterprise_data_roles.models import EnterpriseDataFeatureRole, EnterpriseDataRoleAssignment


@ddt.ddt
@mark.django_db
class TestEnterpriseLearnerEnrollmentViewSet(JWTTestMixin, APITransactionTestCase):
    """
    Tests for EnterpriseLearnerEnrollmentViewSet.
    """

    def setUp(self):
        super().setUp()
        self.user = UserFactory(is_staff=True)
        role, __ = EnterpriseDataFeatureRole.objects.get_or_create(name=ENTERPRISE_DATA_ADMIN_ROLE)
        self.role_assignment = EnterpriseDataRoleAssignment.objects.create(
            role=role,
            user=self.user
        )
        self.client.force_authenticate(user=self.user)

        mocked_get_enterprise_customer = mock.patch(
            'enterprise_data.filters.EnterpriseApiClient.get_enterprise_customer',
            return_value=get_dummy_enterprise_api_data()
        )

        self.mocked_get_enterprise_customer = mocked_get_enterprise_customer.start()
        self.addCleanup(mocked_get_enterprise_customer.stop)
        self.enterprise_id = 'ee5e6b3a-069a-4947-bb8d-d2dbc323396c'
        self.set_jwt_cookie()

    def tearDown(self):
        super().tearDown()
        EnterpriseLearnerEnrollment.objects.all().delete()

    def test_filter_by_offer_id(self):
        enterprise_learner = EnterpriseLearnerFactory(
            enterprise_customer_uuid=self.enterprise_id
        )

        offer_1_id = '1234'
        offer_2_id = '2ThisIsmyOfferId'

        learner_enrollment_1 = EnterpriseLearnerEnrollmentFactory(
            offer_id=offer_1_id,
            enterprise_customer_uuid=self.enterprise_id,
            is_consent_granted=True,
            enterprise_user_id=enterprise_learner.enterprise_user_id
        )
        EnterpriseLearnerEnrollmentFactory(
            offer_id=offer_2_id,
            enterprise_customer_uuid=self.enterprise_id,
            is_consent_granted=True,
            enterprise_user_id=enterprise_learner.enterprise_user_id
        )

        url = reverse('v1:enterprise-learner-enrollment-list', kwargs={'enterprise_id': self.enterprise_id})
        response = self.client.get(url, data={'offer_id': offer_1_id})
        results = response.json()['results']
        self.assertEqual(len(results), 1)
        self.assertEqual(results[0]['enrollment_id'], learner_enrollment_1.enrollment_id)

    def test_filter_by_ignore_null_course_list_price(self):
        enterprise_learner = EnterpriseLearnerFactory(
            enterprise_customer_uuid=self.enterprise_id
        )

        EnterpriseLearnerEnrollmentFactory(
            course_list_price=None,
            enterprise_customer_uuid=self.enterprise_id,
            is_consent_granted=True,
            enterprise_user_id=enterprise_learner.enterprise_user_id
        )
        learner_enrollment_with_price = EnterpriseLearnerEnrollmentFactory(
            enterprise_customer_uuid=self.enterprise_id,
            is_consent_granted=True,
            enterprise_user_id=enterprise_learner.enterprise_user_id
        )

        url = reverse('v1:enterprise-learner-enrollment-list', kwargs={'enterprise_id': self.enterprise_id})
        response = self.client.get(url, data={'ignore_null_course_list_price': True})
        results = response.json()['results']
        self.assertEqual(len(results), 1)
        self.assertEqual(results[0]['enrollment_id'], learner_enrollment_with_price.enrollment_id)

    def test_get_course_product_line(self):
        """ Test that the course product line information is returned correctly """
        enterprise_learner = EnterpriseLearnerFactory(
            enterprise_customer_uuid=self.enterprise_id
        )
        learner_enrollment_executive_ed = EnterpriseLearnerEnrollmentFactory(
            enterprise_customer_uuid=self.enterprise_id,
            is_consent_granted=True,
            enterprise_user_id=enterprise_learner.enterprise_user_id,
            course_product_line='Executive Education'
        )
        EnterpriseLearnerEnrollmentFactory(
            enterprise_customer_uuid=self.enterprise_id,
            is_consent_granted=True,
            enterprise_user_id=enterprise_learner.enterprise_user_id,
            course_product_line='OCM'
        )

        url = reverse('v1:enterprise-learner-enrollment-list', kwargs={'enterprise_id': self.enterprise_id})
        response = self.client.get(url, data={'course_product_line': 'Executive Education'})
        results = response.json()['results']
        self.assertEqual(len(results), 1)
        self.assertEqual(results[0]['enrollment_id'], learner_enrollment_executive_ed.enrollment_id)

    def test_get_subsidy_flag(self):
        """ Test that the subsidy information is returned correctly """
        enterprise_learner = EnterpriseLearnerFactory(
            enterprise_customer_uuid=self.enterprise_id
        )
        learner_enrollment_subsidized = EnterpriseLearnerEnrollmentFactory(
            enterprise_customer_uuid=self.enterprise_id,
            is_consent_granted=True,
            enterprise_user_id=enterprise_learner.enterprise_user_id,
            is_subsidy=True
        )
        EnterpriseLearnerEnrollmentFactory(
            enterprise_customer_uuid=self.enterprise_id,
            is_consent_granted=True,
            enterprise_user_id=enterprise_learner.enterprise_user_id,
            is_subsidy=False
        )

        url = reverse('v1:enterprise-learner-enrollment-list', kwargs={'enterprise_id': self.enterprise_id})
        response = self.client.get(url, data={'is_subsidy': True})
        results = response.json()['results']
        self.assertEqual(len(results), 1)
        self.assertEqual(results[0]['enrollment_id'], learner_enrollment_subsidized.enrollment_id)

    def test_search_by_course_title(self):
        """Test that the course title search works correctly"""
        enterprise_learner = EnterpriseLearnerFactory(
            enterprise_customer_uuid=self.enterprise_id
        )
        course_titles = ['Introduction to Python', 'Introduction to Java', 'Introduction to C++']

        for title in course_titles:
            EnterpriseLearnerEnrollmentFactory(
                enterprise_customer_uuid=self.enterprise_id,
                is_consent_granted=True,
                enterprise_user_id=enterprise_learner.enterprise_user_id,
                course_title=title,
            )

        url = reverse('v1:enterprise-learner-enrollment-list', kwargs={'enterprise_id': self.enterprise_id})
        response = self.client.get(url, data={'search_all': 'Introduction to'})
        results = response.json()['results']
        self.assertEqual(len(results), 3)

    def test_search_by_email(self):
        """Test that the email search works correctly"""
        enterprise_learner = EnterpriseLearnerFactory(
            enterprise_customer_uuid=self.enterprise_id,
            user_email="johndoe@example.com"
        )
        EnterpriseLearnerEnrollmentFactory(
            enterprise_customer_uuid=self.enterprise_id,
            is_consent_granted=True,
            enterprise_user_id=enterprise_learner.enterprise_user_id,
            course_title="Sample course",
            user_email="johndoe@example.com"
        )
        url = reverse('v1:enterprise-learner-enrollment-list', kwargs={'enterprise_id': self.enterprise_id})
        response = self.client.get(url, data={'search_all': 'john'})
        results = response.json()['results']
        self.assertEqual(len(results), 1)

    def test_search_no_results(self):
        """Test that the search returns no results if no matches are found"""
        EnterpriseLearnerFactory(
            enterprise_customer_uuid=self.enterprise_id,
            user_email="test@example.com"
        )
        search_term = 'nonexistentemail@example.com'
        url = reverse('v1:enterprise-learner-enrollment-list', kwargs={'enterprise_id': self.enterprise_id})
        response = self.client.get(url, data={'search_all': search_term})
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.json()['count'], 0)


@ddt.ddt
@mark.django_db
class TestEnterpriseOffersViewSet(JWTTestMixin, APITransactionTestCase):
    """
    Tests for EnterpriseOfferViewSet.
    """

    def setUp(self):
        super().setUp()
        self.user = UserFactory(is_staff=True)
        role, __ = EnterpriseDataFeatureRole.objects.get_or_create(name=ENTERPRISE_DATA_ADMIN_ROLE)
        self.role_assignment = EnterpriseDataRoleAssignment.objects.create(
            role=role,
            user=self.user
        )
        self.client.force_authenticate(user=self.user)

        self.enterprise_customer_uuid_1 = uuid4()
        self.enterprise_offer_1_offer_id = str(uuid4())
        self.enterprise_offer_1 = EnterpriseOfferFactory(
            offer_id=self.enterprise_offer_1_offer_id.replace('-', ''),
            enterprise_customer_uuid=self.enterprise_customer_uuid_1
        )

        self.enterprise_customer_2_uuid = uuid4()
        self.enterprise_offer_2_offer_id = '11111'
        self.enterprise_offer_2 = EnterpriseOfferFactory(
            offer_id=self.enterprise_offer_2_offer_id,
            enterprise_customer_uuid=self.enterprise_customer_2_uuid
        )

        self.set_jwt_cookie()

    def tearDown(self):
        super().tearDown()
        EnterpriseOffer.objects.all().delete()

    def test_list_offers(self):
        enterprise_id = self.enterprise_offer_1.enterprise_customer_uuid
        url = reverse(
            'v1:enterprise-offers-list',
            kwargs={'enterprise_id': enterprise_id}
        )

        response = self.client.get(url)
        assert response.status_code == status.HTTP_200_OK

        expected_results = [EnterpriseOfferSerializer(self.enterprise_offer_1).data]
        response_json = response.json()
        results = response_json['results']
        assert results == expected_results

    def test_retrieve_offers_uuid(self):
        """
        Make sure that EnterpriseOffer objects that store UUID values inside offer_id return a hyphenated UUID.
        """
        enterprise_id = self.enterprise_offer_1.enterprise_customer_uuid
        url = os.path.join(
            reverse(
                'v1:enterprise-offers-list',
                kwargs={'enterprise_id': enterprise_id}
            ),
            self.enterprise_offer_1_offer_id + "/",
        )
        response = self.client.get(url)
        assert response.status_code == status.HTTP_200_OK

        response_json = response.json()
        results = response_json
        assert results['offer_id'] == str(UUID(self.enterprise_offer_1_offer_id))

    def test_retrieve_offer_offer_id_int(self):
        """
        Make sure that EnterpriseOffer objects that store integer values inside offer_id return the value verbatim.
        """
        enterprise_id = self.enterprise_offer_2.enterprise_customer_uuid
        url = os.path.join(
            reverse(
                'v1:enterprise-offers-list',
                kwargs={'enterprise_id': enterprise_id}
            ),
            self.enterprise_offer_2_offer_id + "/",
        )
        response = self.client.get(url)
        assert response.status_code == status.HTTP_200_OK

        response_json = response.json()
        results = response_json
        assert results['offer_id'] == self.enterprise_offer_2_offer_id


@ddt.ddt
@mark.django_db
class TestEnterpriseAdminInsightsView(JWTTestMixin, APITransactionTestCase):
    """
    Tests for EnterpriseAdminInsightsView.
    """

    def setUp(self):
        super().setUp()
        self.user = UserFactory()
        self.enterprise_customer_uuid_1 = uuid4()
        self.enterprise_customer_uuid_2 = uuid4()
        self.set_jwt_cookie(context=f'{self.enterprise_customer_uuid_1}')

    def verify_data(self, received, expected):
        """
        Verify that received and expected data are same.
        """
        received_keys = received.keys()
        for key in received_keys:
            expected_value = getattr(expected, key)
            if isinstance(expected_value, UUID):
                expected_value = str(expected_value)
            if isinstance(expected_value, datetime.datetime):
                expected_value = expected_value.strftime("%Y-%m-%dT%H:%M:%S.%fZ")

            assert received[key] == expected_value

    def test_retrieve_enterprise_admin_insights(self):
        """
        Verify that `EnterpriseAdminInsightsView` gives correct response if the user has access to enterprise.
        """
        enterprise_customer_uuid = self.enterprise_customer_uuid_1
        learner_progress_obj = EnterpriseAdminLearnerProgressFactory.create(
            enterprise_customer_uuid=enterprise_customer_uuid
        )
        learner_engagement_obj = EnterpriseAdminSummarizeInsightsFactory.create(
            enterprise_customer_uuid=enterprise_customer_uuid
        )

        url = reverse('v1:enterprise-admin-insights', kwargs={'enterprise_id': enterprise_customer_uuid})
        response = self.client.get(url)
        assert response.status_code == status.HTTP_200_OK

        response_json = response.json()

        learner_progress = response_json['learner_progress']
        self.verify_data(learner_progress, learner_progress_obj)
        learner_engagement = response_json['learner_engagement']
        self.verify_data(learner_engagement, learner_engagement_obj)

    def test_retrieve_enterprise_admin_insights_no_data(self):
        """
        Verify that `EnterpriseAdminInsightsView` gives correct response if the user has access but no data exists.
        """
        url = reverse('v1:enterprise-admin-insights', kwargs={'enterprise_id': self.enterprise_customer_uuid_1})
        response = self.client.get(url)
        assert response.status_code == status.HTTP_404_NOT_FOUND

    def test_retrieve_enterprise_admin_insights_no_access(self):
        """
        Verify that `EnterpriseAdminInsightsView` give 401 if the user has access to enterprise.
        """
        enterprise_customer_uuid = self.enterprise_customer_uuid_2
        url = reverse('v1:enterprise-admin-insights', kwargs={'enterprise_id': enterprise_customer_uuid})
        response = self.client.get(url)
        assert response.status_code == status.HTTP_401_UNAUTHORIZED



"""
Enterprise_data url configuration.
"""

from django.urls import include, path

app_name = 'enterprise_data'
urlpatterns = [
    path('enterprise/api/', include('enterprise_data.api.urls'),
         name='enterprise_data_api'
         ),
]



"""
Utility functions for Enterprise Reporting.
"""


import datetime
import logging
import os
import re
from collections import OrderedDict
from email.mime.application import MIMEApplication
from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText

import boto3
import pgpy
import pyminizip
import pytz
from cryptography.fernet import Fernet
from fernet_fields.hkdf import derive_fernet_key
from urllib.parse import parse_qs, urlparse

from django.utils.encoding import force_str

LOGGER = logging.getLogger(__name__)

COMPRESSION_LEVEL = 4

FREQUENCY_TYPE_DAILY = 'daily'
FREQUENCY_TYPE_MONTHLY = 'monthly'
FREQUENCY_TYPE_WEEKLY = 'weekly'

AWS_REGION = 'us-east-1'


def compress_and_encrypt(files, password=None, pgp_key=''):
    """
    Given file(s) and a password or a PGP key,
    create a password protected or encrypted compressed file.
    Return the new filename.
    """
    if pgp_key:
        zipfile = _get_compressed_file(files)
        return _get_encrypted_file(zipfile, pgp_key)
    else:
        return _get_compressed_file(files, password)


def _get_encrypted_file(zipfile, pgp_key):
    """
    Given a file and a pgp public key, create an encrypted file. Return the new filename.
    """
    rsa_pub, _ = pgpy.PGPKey.from_blob(pgp_key)
    message = pgpy.PGPMessage.new(zipfile, file=True)
    pgpfile = f'{zipfile}.pgp'
    encrypted_message = rsa_pub.encrypt(message)
    with open(pgpfile, 'wb') as encrypted_file:
        encrypted_file.write(encrypted_message.__bytes__())
    return pgpfile


def _get_compressed_file(files, password=None):
    """
    Given file(s) and a password, create a zip file. Return the new filename.
    """
    multiple_files = len(files) > 1
    # Replace the file extension with `.zip`
    zipfile = re.sub(r'\.(\w+)$', '.zip', files[0].name)
    compression = pyminizip.compress_multiple if multiple_files else pyminizip.compress
    src_file_path_prefix = [] if multiple_files else None
    compression(
        [f.name for f in files] if multiple_files else files[0].name,
        src_file_path_prefix,
        zipfile,
        password,
        COMPRESSION_LEVEL
    )
    return zipfile


def prepare_attachments(attachment_data):
    """
    Helper function to create a list contain file attachment objects
    for use with MIMEMultipart

    attachment_data should be a dictionary, with filenames as keys, and None
        or a string as value. If the value is None, the key is assumed to be a
        filename

    attachment_data = {
        'my_file_name.csv': '1,1,1,1,1,\n2,2,2,2,2\n3,3,3,3,3\n'
        '/abs/path/to/myfile.csv': None
    }

    Returns a list of MIMEApplication objects
    """

    attachments = []
    for filename, data in list(attachment_data.items()):
        if data is None:
            msg_attachment = MIMEApplication(open(filename, 'rb').read())
        else:
            msg_attachment = MIMEApplication(data)
        msg_attachment.add_header(
            'Content-Disposition',
            'attachment',
            filename=os.path.basename(filename)
        )
        msg_attachment.set_type('application/zip')
        attachments.append(msg_attachment)

    return attachments


def send_email_with_attachment(subject, body, from_email, to_email, attachment_data):
    """
    Send an email with a file attachment.

    Adapted from https://gist.github.com/yosemitebandit/2883593

    attachment_data should be a dict of file name and data key-value pairs
    """
    # connect to SES
    client = boto3.client('ses', region_name=AWS_REGION)

    # the message body
    msg_body = MIMEText(body)

    # the attachment
    attachments = prepare_attachments(attachment_data)

    # iterate over each email in the list to send emails independently
    for email in to_email:
        msg = MIMEMultipart()
        msg['Subject'] = subject
        msg['From'] = from_email
        msg['To'] = email

        # what a recipient sees if they don't use an email reader
        msg.preamble = 'Multipart message.\n'

        # attach the message body and attachment
        msg.attach(msg_body)
        for msg_attachment in attachments:
            msg.attach(msg_attachment)

        # and send the message
        result = client.send_raw_email(RawMessage={'Data': msg.as_string()}, Source=msg['From'], Destinations=[email])
        LOGGER.debug(result)


def is_current_time_in_schedule(frequency, hour_of_day, day_of_month=None, day_of_week=None):
    """
    Determine if the current time is in the range specified by this configuration's schedule.
    """
    est_timezone = pytz.timezone('US/Eastern')
    current_est_time = datetime.datetime.now(est_timezone)
    current_hour_of_day = current_est_time.hour
    current_day_of_week = current_est_time.weekday()
    current_day_of_month = current_est_time.day

    LOGGER.info(f'Job Current EST: [{current_hour_of_day}-{current_day_of_week}-{current_day_of_month}]')
    LOGGER.info(f'Enterprise Report Schedule: [{frequency}-{hour_of_day}-{day_of_week}-{day_of_month}]')

    # All configurations have an hour of the day, so the hour must always match in order to send a report.
    if hour_of_day == current_hour_of_day:
        # If reports should be sent monthly and today is the same as the day configured, return True
        if frequency == FREQUENCY_TYPE_MONTHLY and day_of_month == current_day_of_month:
            return True
        # If reports should be sent weekly and today is the same as the day configured, return True
        elif frequency == FREQUENCY_TYPE_WEEKLY and day_of_week == current_day_of_week:
            return True
        # If reports should be sent daily, return True
        elif frequency == FREQUENCY_TYPE_DAILY:
            return True

    return False


def encrypt_string(string):
    """
    Encrypts a string using Fernet symmetric encryption.
    """
    fernet = Fernet(
        derive_fernet_key(
            os.environ.get('LMS_FERNET_KEY')
        )
    )

    return force_str(fernet.encrypt(bytes(string, 'utf-8')))


def decrypt_string(string):
    """
    Decrypts a string that was encrypted using Fernet symmetric encryption.
    """
    fernet = Fernet(
        derive_fernet_key(
            os.environ.get('LMS_FERNET_KEY')
        )
    )

    return force_str(fernet.decrypt(bytes(string, 'utf-8')))


def flatten_dict(d, target='key' or 'value'):
    """
    Flattens a dict object into a list.

    The behavior is as such for targeting keys:
        * Each key is concatenated into the list.
        * Each key with a list value has the list's indices formatted (see [1]) and concatenated into the list.
        * Each key with a dict value has the dict's sub-keys formatted (see [1]) and concatenated into the list.
        * Sub-objects are recursively flattened.

    The behavior is as such for targeting values:
        * Each non-list value is concatenated into the list.
        * Each list value has its entries concatenated into the list.
        * Sub-objects are recursively flattened.

    WARNING: Appropriately flattening lists with a mixture of dict and non-dict objects within them is unsupported.

    [1]: The nested-value formatting function can be found nested within this function. (No pun intended).

    # TODO: This beastly mess of a house of cards *needs* a refactor at some point. Make do with comments for now.
    # TODO: This can probably be separated into a dedicated Python library somewhere sometime in the future.
    """
    def format_nested(nested, _key=None):
        if _key is None:
            _key = key
        return f'{_key}_{nested}'

    flattened = []
    target_is_key = target == 'key'
    for key, value in list(OrderedDict(sorted(d.items())).items()):

        # Simple case: recursively flatten the dictionary.
        if isinstance(value, dict):
            flattened += list(map(
                format_nested if target_is_key else lambda x: x,
                flatten_dict(value, target=target)
            ))

        # We are suddenly in muddy waters, because lists can have multiple types within them in JSON.
        elif isinstance(value, list):
            items_are_dict = [isinstance(item, dict) for item in value]
            items_are_list = [isinstance(item, list) for item in value]

            # To help reduce the complexity here, let's not support this case.
            # Besides, most sensible APIs won't bump into this case.
            if any(items_are_dict) and not all(items_are_dict):
                raise NotImplementedError("Ability to flatten dict with list of mixed dict and non-dict types "
                                          "is not currently supported")

            # Same here, this is just weird.
            if any(items_are_list):
                raise NotImplementedError("Ability to flatten a dict with lists within lists "
                                          "is not currently supported. And we'd like to ask you to take it easy.")

            # This case is common, but a little complex.
            elif all(items_are_dict):
                for index, item in enumerate(value):
                    _flattened_dict = flatten_dict(item, target=target)

                    # In this case we actually want to prepend the dict's index in the list to each flattened dict.
                    if target_is_key:
                        _flattened_dict = [format_nested(flattened_item, _key=index)
                                           for flattened_item in _flattened_dict]

                    flattened += list(map(format_nested if target_is_key else lambda x: x, _flattened_dict))

            # All items are non-dict, so just directly add either the index or the value.
            else:
                flattened += list(map(format_nested, list(range(len(value))))) if target_is_key else value

        # Kindergarten -- just add to the list.
        else:
            flattened.append(key if target_is_key else value)
    return flattened


def generate_data(item, target='key' or 'value'):
    """
    Either return a list of JSON data objects or
    List of headers depends upon the target.
    """
    data = []
    target_is_key = target == 'key'
    for key, value in list(OrderedDict(sorted(item.items())).items()):
        if target_is_key:
            data.append(key)
            continue

        # For empty list we are just writing an empty string ''.
        if isinstance(value, list) and not len(value):
            value = ''

        data.append(value)

    return data


def get_content_metadata_item_id(content_metadata_item):
    """
    Return the unique identifier given a content metadata item dictionary.
    """
    if content_metadata_item['content_type'] == 'program':
        return content_metadata_item['uuid']
    if content_metadata_item['content_type'] == 'learnerpathway':
        return content_metadata_item['uuid']
    return content_metadata_item['key']


def traverse_pagination(response, endpoint):
    """
    Traverse a paginated API response.

    Extracts and concatenates "results" (list of dict) returned by DRF-powered
    APIs.

    Arguments:
        response (Dict): Current response dict from service API
        endpoint (slumber Resource object): slumber Resource object from edx-rest-api-client

    Returns:
        list of dict.

    """
    results = response.get('results', [])

    next_page = response.get('next')
    while next_page:
        querystring = parse_qs(urlparse(next_page).query, keep_blank_values=True)
        response = endpoint.get(**querystring)
        results += response.get('results', [])
        next_page = response.get('next')

    return results


def extract_catalog_uuids_from_reporting_config(reporting_config):
    """
    Helper method to extract uuids from reporting config

    Returns a dict with 1 key, 'results', whose value is a list of
    dicts containing a key-value pair of 'uuid' and some uuid
    """
    enterprise_customer_catalogs = {'results': [
        {'uuid': catalog['uuid']}
        for catalog in reporting_config.get('enterprise_customer_catalogs', [])
        ]
    }
    return enterprise_customer_catalogs


def retry_on_exception(max_retries=3, delay=2, backoff=2):
    """
    Decorator to retry a function on exception with exponential backoff. The
    function will be retried on exception up to max_retries times. The delay
    between retries will increase exponentially with each retry.

    :param max_retries: maximum number of retries
    :param delay: initial delay in seconds
    :param backoff: backoff multiplier e.g. value of 2 will double the delay

    :return: decorator
    """

    def backoff_time(retry_count):
        """
        Calculate wait time using exponential backoff formula

        :param retry_count: number of retries
        :return: wait time
        """
        return delay * (backoff ** (retry_count - 1))

    def wait(wait_time):
        """
        Wait for wait_time seconds using time.sleep

        :param wait_time: time to wait in seconds
        :return: None
        """
        import time
        time.sleep(wait_time)

    def decorator_retry(func):
        def wrapper(*args, **kwargs):
            retry_count = 0
            while retry_count < max_retries:  # retry loop
                try:
                    return func(*args, **kwargs)
                except Exception as e:
                    LOGGER.info(
                        'Exception raised while calling %s. Retrying: %s',
                        func.__name__,
                        str(e),
                    )
                    retry_count += 1
                    time = backoff_time(retry_count)
                    wait(time)
            return func(*args, **kwargs)  # last attempt (don't wait)
        return wrapper
    return decorator_retry



# Generated by Django 3.2.21 on 2023-09-22 10:31

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('enterprise_data', '0036_enterprisesubsidybudget_subsidy_access_policy_display_name'),
    ]

    operations = [
        migrations.AlterField(
            model_name='enterpriseenrollment',
            name='consent_granted',
            field=models.BooleanField(default=None, null=True),
        ),
    ]



# Generated by Django 1.11.5 on 2018-06-12 05:34


from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('enterprise_data', '0006_auto_20180612_0336'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='enterpriseenrollment',
            name='last_activity_timestamp',
        ),
        migrations.AddField(
            model_name='enterpriseenrollment',
            name='last_activity_date',
            field=models.DateField(null=True),
        ),
    ]



# Generated by Django 3.2.20 on 2023-09-07 08:34

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('enterprise_data', '0033_enterpriseadminlearnerprogress_enterpriseadminsummarizeinsights'),
    ]

    operations = [
        migrations.CreateModel(
            name='EnterpriseSubsidyBudget',
            fields=[
                ('id', models.CharField(db_index=True, max_length=32, primary_key=True, serialize=False)),
                ('subsidy_access_policy_uuid', models.UUIDField()),
                ('subsidy_uuid', models.UUIDField()),
                ('enterprise_customer_uuid', models.UUIDField()),
                ('enterprise_customer_name', models.CharField(max_length=255, null=True)),
                ('subsidy_access_policy_description', models.CharField(max_length=500, null=True)),
                ('subsidy_title', models.CharField(max_length=255, null=True)),
                ('catalog_name', models.CharField(max_length=255, null=True)),
                ('subsidy_access_policy_type', models.CharField(max_length=255, null=True)),
                ('date_created', models.DateTimeField(null=True)),
                ('subsidy_start_datetime', models.DateTimeField(null=True)),
                ('subsidy_expiration_datetime', models.DateTimeField(null=True)),
                ('is_active', models.BooleanField(default=False)),
                ('is_test', models.BooleanField(default=False)),
                ('ocm_usage', models.FloatField(null=True)),
                ('exec_ed_usage', models.FloatField(null=True)),
                ('usage_total', models.FloatField(null=True)),
                ('enterprise_contract_discount_percent', models.FloatField(null=True)),
                ('starting_balance', models.FloatField(null=True)),
                ('amount_of_policy_spent', models.FloatField(null=True)),
                ('remaining_balance', models.FloatField(null=True)),
                ('percent_of_policy_spent', models.FloatField(null=True)),
            ],
            options={
                'verbose_name': 'Enterprise Subsidy Budget',
                'verbose_name_plural': 'Enterprise Subsidy Budgets',
                'db_table': 'ent_subsidy_access_policy_aggregates',
            },
        ),
        migrations.AddField(
            model_name='enterpriselearnerenrollment',
            name='budget_id',
            field=models.UUIDField(db_index=True, null=True),
        ),
    ]



"""
management command for creating enterprise enrollments in bulk
"""

from django.core.management.base import BaseCommand, CommandError

from enterprise_data.tests.test_utils import EnterpriseEnrollmentFactory, EnterpriseUserFactory


class Command(BaseCommand):
    """ management command class for creating enterprise enrollments in bulk """
    help = 'Creates 10 enterprise users with 5 random enrollments each'

    def add_arguments(self, parser):
        parser.add_argument(
            'enterprise_id',
            type=str,
            help='UUID for an enterprise'
        )

    def handle(self, *args, **options):
        enterprise_id = options['enterprise_id']
        try:
            for _ in range(10):
                user = EnterpriseUserFactory(enterprise_id=enterprise_id)
                for _ in range(5):
                    EnterpriseEnrollmentFactory(
                        enterprise_id=enterprise_id,
                        enterprise_user=user,
                        user_email=user.user_email)
        except Exception as exc:
            info = (
                'Error trying to create Enrollments for enterprise '
                '{}: {}'.format(enterprise_id, exc)
            )
            raise CommandError(info) from exc



# Generated by Django 2.2.17 on 2021-06-02 10:46

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('enterprise_data', '0022_remove_role_based_access_control_switch'),
    ]

    operations = [
        migrations.CreateModel(
            name='EnterpriseLearner',
            fields=[
                ('enterprise_user_id', models.PositiveIntegerField(primary_key=True, serialize=False)),
                ('enterprise_customer_uuid', models.UUIDField(db_index=True)),
                ('enterprise_user_created', models.DateTimeField(null=True)),
                ('enterprise_user_modified', models.DateTimeField(null=True)),
                ('enterprise_user_active', models.BooleanField(default=False)),
                ('lms_user_id', models.PositiveIntegerField()),
                ('is_linked', models.BooleanField(default=False)),
                ('user_username', models.CharField(max_length=255, null=True)),
                ('user_email', models.CharField(db_index=True, max_length=255, null=True)),
                ('lms_user_created', models.DateTimeField(null=True)),
                ('lms_last_login', models.DateTimeField(null=True)),
                ('lms_user_country', models.CharField(max_length=2, null=True)),
                ('enterprise_sso_uid', models.CharField(max_length=255, null=True)),
                ('last_activity_date', models.DateField(db_index=True, null=True)),
                ('created_at', models.DateTimeField(db_index=True, null=True)),
            ],
            options={
                'verbose_name': 'Enterprise Learner',
                'verbose_name_plural': 'Enterprise Learner',
                'db_table': 'enterprise_learner',
            },
        ),
        migrations.CreateModel(
            name='EnterpriseLearnerEnrollment',
            fields=[
                ('enterprise_enrollment_id', models.PositiveIntegerField(primary_key=True, serialize=False)),
                ('enrollment_id', models.PositiveIntegerField(null=True)),
                ('is_consent_granted', models.BooleanField(default=False)),
                ('paid_by', models.CharField(max_length=128, null=True)),
                ('user_current_enrollment_mode', models.CharField(max_length=32)),
                ('enrollment_date', models.DateField()),
                ('unenrollment_date', models.DateField(null=True)),
                ('unenrollment_end_within_date', models.DateField(null=True)),
                ('is_refunded', models.BooleanField(default=None, null=True)),
                ('seat_delivery_method', models.CharField(max_length=128, null=True)),
                ('offer_name', models.CharField(max_length=255, null=True)),
                ('offer_type', models.CharField(max_length=128, null=True)),
                ('coupon_code', models.CharField(max_length=128, null=True)),
                ('coupon_name', models.CharField(max_length=128, null=True)),
                ('contract_id', models.CharField(max_length=128, null=True)),
                ('course_list_price', models.DecimalField(decimal_places=2, max_digits=12, null=True)),
                ('amount_learner_paid', models.DecimalField(decimal_places=2, max_digits=12, null=True)),
                ('course_key', models.CharField(db_index=True, max_length=255)),
                ('courserun_key', models.CharField(db_index=True, max_length=255)),
                ('course_title', models.CharField(db_index=True, max_length=255, null=True)),
                ('course_pacing_type', models.CharField(max_length=32, null=True)),
                ('course_start_date', models.DateField(db_index=True, null=True)),
                ('course_end_date', models.DateField(null=True)),
                ('course_duration_weeks', models.PositiveIntegerField(null=True)),
                ('course_max_effort', models.PositiveIntegerField(null=True)),
                ('course_min_effort', models.PositiveIntegerField(null=True)),
                ('course_primary_program', models.CharField(max_length=128, null=True)),
                ('course_primary_subject', models.CharField(max_length=128, null=True)),
                ('has_passed', models.BooleanField(default=False)),
                ('last_activity_date', models.DateField(db_index=True, null=True)),
                ('progress_status', models.CharField(max_length=128, null=True)),
                ('passed_date', models.DateField(null=True)),
                ('current_grade', models.FloatField(null=True)),
                ('letter_grade', models.CharField(max_length=32, null=True)),
                ('user_email', models.CharField(db_index=True, max_length=255, null=True)),
                ('user_account_creation_date', models.DateTimeField(null=True)),
                ('user_country_code', models.CharField(max_length=2, null=True)),
                ('user_username', models.CharField(max_length=255, null=True)),
                ('enterprise_name', models.CharField(db_index=True, max_length=255)),
                ('enterprise_customer_uuid', models.UUIDField(db_index=True)),
                ('enterprise_sso_uid', models.CharField(max_length=255, null=True)),
                ('created', models.DateTimeField(db_index=True, null=True)),
                ('enterprise_user', models.ForeignKey(null=True, on_delete=django.db.models.deletion.CASCADE, related_name='enrollments', to='enterprise_data.EnterpriseLearner')),
            ],
            options={
                'verbose_name': 'Enterprise Learner Enrollment',
                'verbose_name_plural': 'Enterprise Learner Enrollments',
                'db_table': 'enterprise_learner_enrollment',
            },
        ),
    ]



import os
import tempfile
from zipfile import ZipFile


def create_files(files_data):
    """
    Creates files based on provided file data.
    """
    files = []
    total_size = 0
    for file_data in files_data:
        tf = tempfile.NamedTemporaryFile(suffix='.txt')
        tf.write(file_data['size'] * b'i')
        tf.flush()
        tf.seek(0)

        files.append({
            'file': tf,
            'size': file_data['size'],
        })
        total_size += file_data['size']

    return files, total_size

def verify_compressed(self, delivery_files, files, original_file_size, password):
    """
    Verify that file is compress correctly.
    """
    # Verify file is compressed.
    zip_file_size = os.path.getsize(delivery_files)
    assert zip_file_size < original_file_size

    zipfile = ZipFile(delivery_files, 'r')

    for file in files:
        # Verify text file is present in zip file.
        assert file['file'].name.split('/')[-1] in zipfile.namelist()

        # Verify file content is readable with correct password.
        content = zipfile.read(file['file'].name.split('/')[-1], bytes(password, 'utf-8'))
        assert len(content) == file['size']

        # Also verify file is not accessible with any other passwords i.e wrong passwords.
        with self.assertRaises(RuntimeError):
            zipfile.read(file['file'].name.split('/')[-1], b'wrong-password')



"""
Renderers for enterprise data views.
"""

from rest_framework_csv.renderers import CSVStreamingRenderer


class EnrollmentsCSVRenderer(CSVStreamingRenderer):
    """
    Custom streaming csv renderer for EnterpriseLearnerEnrollment data.
    """

    # This will be used as CSV header for csv generated from `admin-portal`.
    # Do not change the order of fields below. Ordering is important because csv generated
    # on `admin-portal` should match `progress_v3` csv generated in `enterprise_reporting`
    # Order and field names below should match with `EnterpriseLearnerEnrollmentSerializer.fields`
    header = [
        'enrollment_id', 'enterprise_enrollment_id', 'is_consent_granted', 'paid_by',
        'user_current_enrollment_mode', 'enrollment_date', 'unenrollment_date',
        'unenrollment_end_within_date', 'is_refunded', 'seat_delivery_method',
        'offer_id', 'offer_name', 'offer_type', 'coupon_code', 'coupon_name', 'contract_id',
        'course_list_price', 'amount_learner_paid', 'course_key', 'courserun_key',
        'course_title', 'course_pacing_type', 'course_start_date', 'course_end_date',
        'course_duration_weeks', 'course_max_effort', 'course_min_effort',
        'course_primary_program', 'primary_program_type', 'course_primary_subject', 'has_passed',
        'last_activity_date', 'progress_status', 'passed_date', 'current_grade',
        'letter_grade', 'enterprise_user_id', 'user_email', 'user_account_creation_date',
        'user_country_code', 'user_username', 'enterprise_name', 'enterprise_customer_uuid',
        'enterprise_sso_uid', 'created', 'course_api_url', 'total_learning_time_hours', 'is_subsidy',
        'course_product_line', 'budget_id', 'enterprise_group_name', 'enterprise_group_uuid',
    ]



"""
Views for enterprise api v1.
"""

from datetime import date, timedelta
from logging import getLogger
from uuid import UUID

from django_filters.rest_framework import DjangoFilterBackend
from edx_django_utils.cache import TieredCache
from edx_rbac.decorators import permission_required
from edx_rbac.mixins import PermissionRequiredMixin
from edx_rest_framework_extensions.auth.jwt.authentication import JwtAuthentication
from edx_rest_framework_extensions.paginators import DefaultPagination
from rest_framework import filters, viewsets
from rest_framework.decorators import action
from rest_framework.response import Response
from rest_framework.status import HTTP_200_OK, HTTP_404_NOT_FOUND
from rest_framework.views import APIView

from django.conf import settings
from django.core.paginator import Paginator
from django.db.models import Count, Max, OuterRef, Prefetch, Q, Subquery, Value
from django.db.models.fields import IntegerField
from django.db.models.functions import Coalesce
from django.http import StreamingHttpResponse
from django.utils import timezone

from enterprise_data.api.v1 import serializers
from enterprise_data.constants import ANALYTICS_API_VERSION_1
from enterprise_data.filters import AuditEnrollmentsFilterBackend, AuditUsersEnrollmentFilterBackend
from enterprise_data.models import (
    EnterpriseAdminLearnerProgress,
    EnterpriseAdminSummarizeInsights,
    EnterpriseLearner,
    EnterpriseLearnerEnrollment,
    EnterpriseOffer,
)
from enterprise_data.paginators import EnterpriseEnrollmentsPagination
from enterprise_data.renderers import EnrollmentsCSVRenderer
from enterprise_data.utils import get_cache_key

LOGGER = getLogger(__name__)
DEFAULT_LEARNER_CACHE_TIMEOUT = 60 * 10


def subtract_one_month(original_date):
    """
    Returns a date exactly one month prior to the passed in date.
    """
    one_day = timedelta(days=1)
    one_month_earlier = original_date - one_day
    while one_month_earlier.month == original_date.month or one_month_earlier.day > original_date.day:
        one_month_earlier -= one_day
    return one_month_earlier


class EnterpriseViewSetMixin(PermissionRequiredMixin):
    """
    Base class for all Enterprise view sets.
    """
    authentication_classes = (JwtAuthentication,)
    pagination_class = DefaultPagination
    permission_required = 'can_access_enterprise'
    API_VERSION = ANALYTICS_API_VERSION_1

    def paginate_queryset(self, queryset):
        """
        Allows no_page query param to skip pagination
        """
        if 'no_page' in self.request.query_params:
            return None
        return super().paginate_queryset(queryset)


class EnterpriseLearnerEnrollmentViewSet(EnterpriseViewSetMixin, viewsets.ReadOnlyModelViewSet):
    """
    Viewset for routes related to Enterprise course enrollments.
    """
    serializer_class = serializers.EnterpriseLearnerEnrollmentSerializer
    pagination_class = EnterpriseEnrollmentsPagination
    filter_backends = (AuditEnrollmentsFilterBackend, filters.OrderingFilter,)
    ordering_fields = '__all__'
    ordering = ('-last_activity_date',)
    ENROLLMENT_MODE_FILTER = 'user_current_enrollment_mode'
    COUPON_CODE_FILTER = 'coupon_code'
    OFFER_FILTER = 'offer_type'
    # TODO: Remove after we release the streaming csv changes
    # This will be used as CSV header for csv generated from `admin-portal`.
    # Do not change the order of fields below. Ordering is important because csv generated
    # on `admin-portal` should match `progress_v3` csv generated in `enterprise_reporting`
    # Order and field names below should match with `EnterpriseLearnerEnrollmentSerializer.fields`
    header = [
        'enrollment_id', 'enterprise_enrollment_id', 'is_consent_granted', 'paid_by',
        'user_current_enrollment_mode', 'enrollment_date', 'unenrollment_date',
        'unenrollment_end_within_date', 'is_refunded', 'seat_delivery_method',
        'offer_id', 'offer_name', 'offer_type', 'coupon_code', 'coupon_name', 'contract_id',
        'course_list_price', 'amount_learner_paid', 'course_key', 'courserun_key',
        'course_title', 'course_pacing_type', 'course_start_date', 'course_end_date',
        'course_duration_weeks', 'course_max_effort', 'course_min_effort',
        'course_primary_program', 'primary_program_type', 'course_primary_subject', 'has_passed',
        'last_activity_date', 'progress_status', 'passed_date', 'current_grade',
        'letter_grade', 'enterprise_user_id', 'user_email', 'user_account_creation_date',
        'user_country_code', 'user_username', 'enterprise_name', 'enterprise_customer_uuid',
        'enterprise_sso_uid', 'created', 'course_api_url', 'total_learning_time_hours', 'is_subsidy',
        'course_product_line', 'budget_id'
    ]

    # TODO: Remove after we release the streaming csv changes
    def get_renderer_context(self):
        renderer_context = super().get_renderer_context()
        renderer_context['header'] = self.header
        return renderer_context

    def get_queryset(self):
        """
        Returns all learner enrollment records for a given enterprise.
        """
        if getattr(self, 'swagger_fake_view', False):
            # queryset just for schema generation metadata
            # See: https://github.com/axnsan12/drf-yasg/issues/333
            return EnterpriseLearnerEnrollment.objects.none()

        enterprise_customer_uuid = self.kwargs['enterprise_id']
        cache_key = get_cache_key(
            resource='enterprise-learner',
            enterprise_customer=enterprise_customer_uuid,
        )
        cached_response = TieredCache.get_cached_response(cache_key)
        if cached_response.is_found:
            return cached_response.value
        else:
            enrollments = EnterpriseLearnerEnrollment.objects.filter(enterprise_customer_uuid=enterprise_customer_uuid)
            enrollments = self.apply_filters(enrollments)
            TieredCache.set_all_tiers(cache_key, enrollments, DEFAULT_LEARNER_CACHE_TIMEOUT)
            return enrollments

    def list(self, request, *args, **kwargs):
        """
        Override the list method to handle streaming CSV download.
        """
        if request.accepted_renderer.format == 'csv':
            return StreamingHttpResponse(
                EnrollmentsCSVRenderer().render(self._stream_serialized_data()),
                content_type="text/csv",
                headers={"Content-Disposition": 'attachment; filename="learner_progress_report.csv"'},
            )

        return super().list(request, *args, **kwargs)

    def _stream_serialized_data(self):
        """
        Stream the serialized data.
        """
        queryset = self.filter_queryset(self.get_queryset())
        serializer = self.get_serializer_class()
        paginator = Paginator(queryset, per_page=settings.ENROLLMENTS_PAGE_SIZE)
        for page_number in paginator.page_range:
            yield from serializer(paginator.page(page_number).object_list, many=True).data

    def apply_filters(self, queryset):
        """
        Filters enrollments based on query params.
        """
        query_filters = self.request.query_params

        past_week_date = date.today() - timedelta(weeks=1)
        past_month_date = subtract_one_month(date.today())

        passed_date_param = query_filters.get('passed_date')
        if passed_date_param == 'last_week':
            queryset = self.filter_past_week_completions(queryset)

        learner_activity_param = query_filters.get('learner_activity')
        if learner_activity_param:
            queryset = self.filter_active_enrollments(queryset)
        if learner_activity_param == 'active_past_week':
            queryset = self.filter_active_learners(queryset, past_week_date)
        elif learner_activity_param == 'inactive_past_week':
            queryset = self.filter_inactive_learners(queryset, past_week_date)
        elif learner_activity_param == 'inactive_past_month':
            queryset = self.filter_inactive_learners(queryset, past_month_date)

        search_email = query_filters.get('search')
        if search_email:
            queryset = queryset.filter(user_email__icontains=search_email)
        search_all_param = query_filters.get('search_all')
        if search_all_param:
            queryset = queryset.filter(
                Q(user_email__icontains=search_all_param)
                | Q(course_title__icontains=search_all_param)
            )
        search_course = query_filters.get('search_course')
        if search_course:
            queryset = queryset.filter(course_title__icontains=search_course)
        search_start_date = query_filters.get('search_start_date')
        if search_start_date:
            queryset = queryset.filter(course_start_date=search_start_date)

        offer_id = query_filters.get('offer_id')
        if offer_id:
            queryset = self.filter_by_offer_id(queryset, offer_id)

        budget_id = query_filters.get('budget_id')
        if budget_id:
            queryset = queryset.filter(budget_id=budget_id)

        ignore_null_course_list_price = query_filters.get('ignore_null_course_list_price')
        if ignore_null_course_list_price:
            queryset = queryset.filter(course_list_price__isnull=False)

        course_product_line = query_filters.get('course_product_line')
        if course_product_line:
            queryset = queryset.filter(course_product_line=course_product_line)

        is_subsidy = query_filters.get('is_subsidy')
        if is_subsidy:
            queryset = queryset.filter(is_subsidy=is_subsidy)

        return queryset

    def filter_by_offer_id(self, queryset, offer_id):
        """
        Filter queryset by `offer_id`
        """
        # offer_id field in table contains two types of data
        # 1. integer
        # 2. uuid
        # In table, uuids are stored without hyphen and in lower case
        # but api clients are sending uuids with hyphen and that causes the issue
        # we need to remove the hyphens from uuid, convert the uuid to lower case and then do a compare in DB

        # if offer_id is a uuid only then do the transformation
        try:
            offer_id_uuid_obj = UUID(offer_id, version=4)
            offer_id = str(offer_id_uuid_obj)
            offer_id = offer_id.replace('-', '')
        except ValueError:
            pass

        return queryset.filter(offer_id=offer_id)

    def filter_active_enrollments(self, queryset):
        """
        Filters queryset to include enrollments with course date in future
        and learners have not passed the course yet.
        """
        return queryset.filter(
            has_passed=False,
            course_end_date__gte=timezone.now(),
        )

    def filter_distinct_learners(self, queryset):
        """
        Filters queryset to include enrollments with a distinct `enterprise_user_id`.
        """
        return queryset.values_list('enterprise_user_id', flat=True).distinct()

    def filter_active_learners(self, queryset, last_activity_date):
        """
        Filters queryset to include enrollments more recent than the specified `last_activity_date`.
        """
        return queryset.filter(last_activity_date__gte=last_activity_date)

    def filter_inactive_learners(self, queryset, last_activity_date):
        """
        Filters queryset to include enrollments more older than the specified `last_activity_date`.
        """
        return queryset.filter(last_activity_date__lte=last_activity_date)

    def filter_distinct_active_learners(self, queryset, last_activity_date):
        """
        Filters queryset to include distinct enrollments more recent than the specified `last_activity_date`.
        """
        return self.filter_distinct_learners(self.filter_active_learners(queryset, last_activity_date))

    def filter_course_completions(self, queryset):
        """
        Filters queryset to include only enrollments that are passing (i.e., completion).
        """
        return queryset.filter(has_passed=1)

    def filter_past_week_completions(self, queryset):
        """
        Filters only those learners who completed a course in last week.
        """
        date_today = timezone.now()
        date_week_before = date_today - timedelta(weeks=1)
        return queryset.filter(
            has_passed=True,
            passed_date__lte=date_today,
            passed_date__gte=date_week_before,
        )

    def filter_number_of_users(self):
        """
        Returns number of enterprise users (enrolled AND not enrolled learners)
        """
        return EnterpriseLearner.objects.filter(enterprise_customer_uuid=self.kwargs['enterprise_id'])

    def get_max_created_date(self, queryset):
        """
        Gets the maximum created timestamp from the queryset.
        """
        created_max = queryset.aggregate(Max('created'))
        return created_max['created__max']

    @action(detail=False)
    def overview(self, request, **kwargs):
        """
        Returns the following data:
            - # of enrolled learners;
            - # of active learners in the past week/month;
            - # of course completions.
            - # of enterprise users (LMS)
        """
        enrollments = self.get_queryset()
        enrollments = self.filter_queryset(enrollments)
        course_completions = self.filter_course_completions(enrollments)
        distinct_learners = self.filter_distinct_learners(enrollments)
        past_week_date = date.today() - timedelta(weeks=1)
        past_month_date = subtract_one_month(date.today())
        active_learners_week = self.filter_distinct_active_learners(enrollments, past_week_date)
        last_updated_date = self.get_max_created_date(enrollments)
        active_learners_month = self.filter_distinct_active_learners(enrollments, past_month_date)
        enterprise_users = self.filter_number_of_users()
        content = {
            'enrolled_learners': distinct_learners.count(),
            'active_learners': {
                'past_week': active_learners_week.count(),
                'past_month': active_learners_month.count(),
            },
            'course_completions': course_completions.count(),
            'last_updated_date': last_updated_date,
            'number_of_users': enterprise_users.count(),
        }
        return Response(content)


class EnterpriseOfferViewSet(EnterpriseViewSetMixin, viewsets.ReadOnlyModelViewSet):
    """
    Viewset for enterprise offers.
    """
    serializer_class = serializers.EnterpriseOfferSerializer
    filter_backends = (filters.OrderingFilter, DjangoFilterBackend,)
    ordering_fields = '__all__'

    lookup_field = 'offer_id'

    filterset_fields = (
        'offer_id',
        'status'
    )

    def get_object(self):
        """
        This ensures that UUIDs with dashes are properly handled when requesting info about offers.

        Related to the work in EnterpriseOfferSerializer with `to_internal_value` and `to_representation`
        """
        self.kwargs['offer_id'] = self.kwargs['offer_id'].replace('-', '')
        return super().get_object()

    def get_queryset(self):
        enterprise_customer_uuid = self.kwargs['enterprise_id']
        return EnterpriseOffer.objects.filter(
            enterprise_customer_uuid=enterprise_customer_uuid,
        )


class EnterpriseLearnerViewSet(EnterpriseViewSetMixin, viewsets.ReadOnlyModelViewSet):
    """
    Viewset for routes related to Enterprise Learners.
    """
    queryset = EnterpriseLearner.objects.all()
    serializer_class = serializers.EnterpriseLearnerSerializer
    filter_backends = (filters.OrderingFilter, AuditUsersEnrollmentFilterBackend,)
    ordering_fields = '__all__'
    ordering = ('user_email',)

    def get_queryset(self):
        LOGGER.info("[ELV_ANALYTICS_API_V1] QueryParams: [%s]", self.request.query_params)
        queryset = super().get_queryset().prefetch_related(
            Prefetch(
                'enrollments',
                queryset=EnterpriseLearnerEnrollment.objects.filter(
                    enterprise_customer_uuid=self.kwargs['enterprise_id']
                )
            ),
        )

        has_enrollments = self.request.query_params.get('has_enrollments')
        if has_enrollments == 'true':
            queryset = queryset.filter(enrollments__isnull=False).distinct()
        elif has_enrollments == 'false':
            queryset = queryset.filter(enrollments__isnull=True)

        active_courses = self.request.query_params.get('active_courses')
        if active_courses == 'true':
            queryset = queryset.filter(
                Q(enrollments__is_consent_granted=True),
                enrollments__course_end_date__gte=timezone.now()
            )
        elif active_courses == 'false':
            queryset = queryset.filter(
                Q(enrollments__is_consent_granted=True),
                enrollments__course_end_date__lte=timezone.now()
            )

        all_enrollments_passed = self.request.query_params.get('all_enrollments_passed')
        if all_enrollments_passed == 'true':
            queryset = queryset.filter(enrollments__has_passed=True)
        elif all_enrollments_passed == 'false':
            queryset = queryset.filter(enrollments__has_passed=False)

        extra_fields = self.request.query_params.getlist('extra_fields')
        if 'enrollment_count' in extra_fields:
            enrollment_subquery = (
                EnterpriseLearnerEnrollment.objects.filter(
                    enterprise_user=OuterRef("enterprise_user_id"),
                    is_consent_granted=True,
                )
                .values('enterprise_user')
                .annotate(enrollment_count=Count('pk', distinct=True))
                .values('enrollment_count')
            )
            queryset = queryset.annotate(
                enrollment_count=Coalesce(
                    Subquery(enrollment_subquery, output_field=IntegerField()),
                    Value(0),
                )
            )

        # based on https://stackoverflow.com/questions/43770118/simple-subquery-with-outerref
        if 'course_completion_count' in extra_fields:
            enrollment_subquery = (
                EnterpriseLearnerEnrollment.objects.filter(
                    enterprise_user=OuterRef("enterprise_user_id"),
                    has_passed=True,
                    is_consent_granted=True,
                )
                .values('enterprise_user')
                .annotate(course_completion_count=Count('pk', distinct=True))
                .values('course_completion_count')
            )
            # Coalesce and Value used here so we don't return "null" to the
            # frontend if the count is 0
            queryset = queryset.annotate(
                course_completion_count=Coalesce(
                    Subquery(enrollment_subquery, output_field=IntegerField()),
                    Value(0),
                )
            )

        return queryset

    def list(self, request, **kwargs):  # pylint: disable=arguments-differ
        """
        List view for learner records for a given enterprise.
        """
        users = self.get_queryset().filter(enterprise_customer_uuid=kwargs['enterprise_id'])

        # do the sorting
        users = self.filter_queryset(users)
        LOGGER.info("[ELV_ANALYTICS_API_V1] QuerySet.Query: [%s]", users.query)

        # Bit to account for pagination
        page = self.paginate_queryset(users)
        if page is not None:
            serializer = self.get_serializer(page, many=True)
            return self.get_paginated_response(serializer.data)
        serializer = self.get_serializer(users, many=True)
        return Response(serializer.data)


class EnterpriseLearnerCompletedCoursesViewSet(EnterpriseViewSetMixin, viewsets.ReadOnlyModelViewSet):
    """
    View to manage enterprise learner completed course enrollments.
    """
    serializer_class = serializers.LearnerCompletedCoursesSerializer
    filter_backends = (filters.OrderingFilter,)
    ordering_fields = '__all__'
    ordering = ('user_email',)

    def get_queryset(self):
        """
        Returns number of completed courses against each learner.
        """
        if getattr(self, 'swagger_fake_view', False):
            # queryset just for schema generation metadata
            # See: https://github.com/axnsan12/drf-yasg/issues/333
            return EnterpriseLearnerEnrollment.objects.none()

        # Get the number of completed courses against a learner.
        enrollments = EnterpriseLearnerEnrollment.objects.filter(
            enterprise_customer_uuid=self.kwargs['enterprise_id'],
            has_passed=True,
            is_consent_granted=True,  # DSC check required
        ).values('user_email').annotate(completed_courses=Count('courserun_key')).order_by('user_email')
        return enrollments


class EnterpriseAdminInsightsView(APIView):
    """
    API for getting the enterprise admin insights.
    """
    authentication_classes = (JwtAuthentication,)
    http_method_names = ['get']

    @permission_required('can_access_enterprise', fn=lambda request, enterprise_id: enterprise_id)
    def get(self, request, enterprise_id):
        """
        HTTP GET endpoint to retrieve the enterprise admin insights
        """
        response_data = {}
        learner_progress = {}
        learner_engagement = {}

        try:
            learner_progress = EnterpriseAdminLearnerProgress.objects.get(enterprise_customer_uuid=enterprise_id)
            learner_progress = serializers.EnterpriseAdminLearnerProgressSerializer(learner_progress).data
            response_data['learner_progress'] = learner_progress
        except EnterpriseAdminLearnerProgress.DoesNotExist:
            pass

        try:
            learner_engagement = EnterpriseAdminSummarizeInsights.objects.get(enterprise_customer_uuid=enterprise_id)
            learner_engagement = serializers.EnterpriseAdminSummarizeInsightsSerializer(learner_engagement).data
            response_data['learner_engagement'] = learner_engagement
        except EnterpriseAdminSummarizeInsights.DoesNotExist:
            pass

        status = HTTP_200_OK
        if learner_progress == {} and learner_engagement == {}:
            status = HTTP_404_NOT_FOUND

        return Response(data=response_data, status=status)
